/*************************************************************************
* (C) 2010 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: aitrpc.c,v 1.20 2015/07/22 20:01:46 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


#pragma GCC visibility push(hidden)

int rpc_Errno;
char rpc_Error[STRSIZ];

#pragma GCC visibility pop

// rpc_GetErrno() Get error code of last operation
int
rpc_GetErrno()
{
	return rpc_Errno;
}

// rpc_GetError() Get error text of last operation
const char *
rpc_GetError()
{
	return rpc_Error;
}

// rpc_SetErr() Set error to variables for internal use!!!
void
rpc_SetErr(int eno, char *estr, ...)
{
	va_list lst;

	rpc_Errno = eno;
	memset(rpc_Error, 0, sizeof rpc_Error);
	va_start(lst, estr);
	vsnprintf(rpc_Error, sizeof rpc_Error, estr, lst);
	va_end(lst);
}


/*
 * rpc_chkPktSession() - Check RPC session
 *
 * @p = packet session
 * @s = active session
 * return: -1, 1, 2, 3 are errors or 0 ok
 */
int
rpc_chkPktSession(rpc_sess_t *p, rpc_sess_t *s)
{
	if (!p || !s)
		return -1;

	if (p->sess_version != s->sess_version)
		return 1;
	if (p->sess_instance != s->sess_instance)
		return 2;

	return 0;
}

/*
 * rpc_addPktSession() - Prepare session into network format
 *
 * @p = packet session
 * @s = host session
 * return: -1 error or 0 ok
 */
int
rpc_addPktSession(rpc_sess_t *p, rpc_sess_t *s)
{
	if (!p || !s)
		return -1;

	p->sess_version = s->sess_version;
	p->sess_instance = s->sess_instance;

	return 0;
}

/*
 * rpc_Read() - RPC read operation
 *
 * @sock = socket
 * @type = type of socket
 * @flags = receive flags
 * @sa = check client address, if you use udp protocol
 * @pkt = RPC packet
 * return: -1 error, 0 EOF or or >0 readed bytes into buffer
 */
ssize_t
rpc_Read(int sock, int type, int flags, sockaddr_t * __restrict sa, ait_val_t * __restrict pkt)
{
	struct pollfd pfd;
	sockaddr_t sa2;
	socklen_t salen;
	int ret = 0, hlen, cx = 0;
	struct ether_header *eh;
	struct bpf_hdr *h;
	ether_addr_t bcst = {{ [0 ... sizeof bcst - 1] = 0xff }};
	u_char *buf = AIT_GET_BUF(pkt);
	size_t blen = AIT_LEN(pkt);
	struct tagRPCCall *rpc = (struct tagRPCCall *) buf;
	ait_val_t v = AIT_VAL_INIT;

try2read:
	pfd.fd = sock;
	pfd.events = POLLIN | POLLPRI;
	memset(buf, 0, blen);
	memset(&sa2, 0, sizeof sa2);
	salen = sa2.ss.ss_len = (u_char) MIN(sizeof(sockaddr_t), 0xff);
	if ((ret = poll(&pfd, 1, DEF_RPC_TIMEOUT * 1000)) < 1 || 
			pfd.revents & (POLLERR | POLLHUP | POLLNVAL)) {
		if (ret)
			LOGERR;
		else
			rpc_SetErr(ETIMEDOUT, "Timeout reached! Server not respond");
		return -1;
	}

	switch (type) {
		case SOCK_STREAM:
			ret = recv(sock, buf, blen, flags);
			break;
		case SOCK_EXT:
			ret = read(sock, buf, blen);
			break;
		case SOCK_BPF:
			if (!sa) {
				rpc_SetErr(EINVAL, "Invalid argument(s)!");
				return -1;
			}

			AIT_SET_BUF(&v, NULL, blen);
			h = (struct bpf_hdr*) AIT_GET_BUF(&v);

			ret = read(sock, AIT_GET_BUF(&v), AIT_LEN(&v));
			if (ret > 0) {
				ret -= h->bh_hdrlen;
				if (ret < h->bh_caplen || h->bh_caplen != h->bh_datalen || 
						ret < ETHER_HDR_LEN + sizeof(struct tagRPCCall)) {
					AIT_FREE_VAL(&v);
					if (cx < 3) {
						cx++;
						ret ^= ret;
						goto try2read;	/* wait for known address */
					} else
						return -1;
				}
				ret = h->bh_caplen;
				eh = (struct ether_header*) (AIT_GET_BUF(&v) + h->bh_hdrlen);
				ret -= ETHER_HDR_LEN;
				if (eh->ether_type != ntohs(RPC_DEFPORT)) {
					AIT_FREE_VAL(&v);
					if (cx < 3) {
						cx++;
						ret ^= ret;
						goto try2read;	/* wait for known address */
					} else
						return -1;
				}

				if (!memcmp(bcst.octet, eh->ether_dhost, sizeof bcst) || 
						!memcmp(bcst.octet, eh->ether_shost, sizeof bcst)) {
					AIT_FREE_VAL(&v);
					if (cx < 3) {
						cx++;
						ret ^= ret;
						goto try2read;	/* wait for known address */
					} else
						return -1;
				}
				memcpy(buf, (u_char*) (eh + 1), MIN(ret, blen));
				AIT_FREE_VAL(&v);
			}
			break;
		case SOCK_RAW:
		case SOCK_DGRAM:
			if (!sa) {
				rpc_SetErr(EINVAL, "Invalid argument(s)!");
				return -1;
			}

			ret = recvfrom(sock, buf, blen, flags, &sa2.sa, &salen);
			if (ret > -1 && e_addrcmp(sa, &sa2, 42)) {
				if (cx < 3) {
					cx++;
					ret ^= ret;
					goto try2read;	/* wait for known address */
				} else
					return -1;
			}
			break;
		default:
			rpc_SetErr(EINVAL, "Invalid argument(s)!");
			return -1;
	}
	if (ret < 0) {
		LOGERR;
		return -1;
	}
	if (!ret)	/* EOF */
		return 0;

	/* check RPC packet header */
	if (type == SOCK_RAW) {
		hlen = sa->sa.sa_family == AF_INET ? 
			sizeof(struct ip) : sizeof(struct ip6_hdr);
		ret -= hlen;
		if (ret > 0)
			memmove(buf, buf + hlen, blen - hlen);
	}

	/* 1st read for RPC header */
	if (ret < sizeof(struct tagRPCCall) || ret < ntohl(rpc->call_len)) {
		rpc_SetErr(ERPCMISMATCH, "Short RPC packet %d bytes", ret);
		return -1;
	}
	/* check for loop request */
	if (!(rpc->call_io & RPC_ACK)) {
		if (cx < 3) {
			cx++;
			ret ^= ret;
			goto try2read;	/* wait for known address */
		} else {
			rpc_SetErr(ERPCMISMATCH, "Loop in RPC communication");
			return -1;
		}
	}

	return ret;
}

/*
 * rpc_Write() - RPC write operation
 *
 * @sock = socket
 * @type = type of socket
 * @flags = send flags
 * @sa = send to client address, if you use udp protocol
 * @pkt = RPC packet
 * @blen = buffer length
 * return: -1 error, 0 EOF or >0 written bytes into buffer
 */
ssize_t
rpc_Write(int sock, int type, int flags, sockaddr_t * __restrict sa, 
		ait_val_t * __restrict pkt, size_t blen)
{
	struct pollfd pfd;
	int ret = 0;
	u_char *buf = AIT_GET_BUF(pkt);
	ait_val_t v = AIT_VAL_INIT;
	struct ether_header *eh;


	pfd.fd = sock;
	pfd.events = POLLOUT;
	if ((ret = poll(&pfd, 1, DEF_RPC_TIMEOUT * 1000)) < 1 || 
			pfd.revents & (POLLERR | POLLHUP | POLLNVAL)) {
		if (ret > 0)
			rpc_SetErr(EPIPE, "Disconnected RPC session\n");
		else
			LOGERR;
		return -1;
	}

	switch (type) {
		case SOCK_STREAM:
			ret = send(sock, buf, blen, flags);
			break;
		case SOCK_EXT:
			ret = write(sock, buf, blen);
			break;
		case SOCK_BPF:
			if (!sa) {
				rpc_SetErr(EINVAL, "Invalid argument(s)!");
				return -1;
			}

			AIT_SET_BUF(&v, NULL, blen + sizeof(struct ether_header));
			eh = (struct ether_header*) AIT_GET_BUF(&v);
			memcpy(eh->ether_dhost, LLADDR(&sa->sdl), ETHER_ADDR_LEN);
			eh->ether_type = htons(RPC_DEFPORT);
			memcpy(eh + 1, buf, blen);
			blen += sizeof(struct ether_header);

			ret = write(sock, AIT_GET_BUF(&v), AIT_LEN(&v));

			AIT_FREE_VAL(&v);
			break;
		case SOCK_RAW:
		case SOCK_DGRAM:
			if (!sa) {
				rpc_SetErr(EINVAL, "Invalid argument(s)!");
				return -1;
			}

			ret = sendto(sock, buf, blen, flags, &sa->sa, sa->sa.sa_len);
			break;
		default:
			rpc_SetErr(EINVAL, "Invalid argument(s)!");
			return -1;
	}
	if (ret < 0) {
		LOGERR;
		return -1;
	}
	if (!ret)	/* EOF */
		return 0;

	if (ret != blen) {
		rpc_SetErr(EPROCUNAVAIL, "RPC request, should be send %d bytes, "
				"really sended %d bytes", blen, ret);
		return -1;
	}

	return ret;
}

/*
 * rpc_pktFreeSpace() - Get free space for payload into RPC packet
 *
 * @c = RPC client 
 * return: remains free bytes from packet
 */
size_t
rpc_pktFreeSpace(rpc_cli_t * __restrict c)
{
	return (sizeof(struct tagRPCCall) + ait_resideVars(RPC_RETVARS(c)));
}
