/*************************************************************************
* (C) 2010 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: blob.c,v 1.17 2015/07/02 22:28:15 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


static void *
toutBLOB(sched_task_t *task)
{
	rpc_srv_unregisterBLOB(TASK_DATA(task), ((rpc_blob_t*) TASK_ARG(task))->blob_var);

	return NULL;
}


/*
 * rpc_srv_blobCreate() - Create and map blob to memory region and return object
 *
 * @srv = RPC Server instance
 * @len = BLOB length object
 * @tout = BLOB live timeout in seconds
 * return: NULL error or !=NULL allocated BLOB object
 */
rpc_blob_t *
rpc_srv_blobCreate(rpc_srv_t * __restrict srv, int len, int tout)
{
	rpc_blob_t *blob = NULL;
	char szFName[MAXPATHLEN];
	int f;
	u_int rnd;
	struct timespec ts = { tout ? tout : DEF_RPC_BLOB_TIMEOUT, 0 };

again:
	rnd = random() % UINT_MAX;

	memset(szFName, 0, sizeof szFName);
	snprintf(szFName, sizeof szFName, BLOB_FILE, AIT_GET_STRZ(&srv->srv_blob.dir), rnd);
	f = open(szFName, O_CREAT | O_EXCL | O_RDWR, 0600);
	if (f == -1) {
		if (errno == EEXIST)
			goto again;

		LOGERR;
		return NULL;
	}
	if (ftruncate(f, len) == -1) {
		LOGERR;
		close(f);
		unlink(szFName);
		return NULL;
	}

	blob = e_malloc(sizeof(rpc_blob_t));
	if (!blob) {
		LOGERR;
		close(f);
		unlink(szFName);
		return NULL;
	}

	blob->blob_data = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, f, 0);
	if (blob->blob_data == MAP_FAILED) {
		LOGERR;
		e_free(blob);
		close(f);
		unlink(szFName);
		return NULL;
	} else {
		close(f);

		madvise(blob->blob_data, len, MADV_SEQUENTIAL);
	}

	blob->blob_len = len;
	blob->blob_var = rnd;

	schedTimer(srv->srv_blob.root, toutBLOB, blob, ts, srv, 0);
	return blob;
}

/*
 * rpc_srv_blobMap() - Map blob to memory region 
 *
 * @srv = RPC Server instance
 * @blob = Map to this BLOB element
 * return: -1 error or 0 ok
 */
int
rpc_srv_blobMap(rpc_srv_t * __restrict srv, rpc_blob_t * __restrict blob)
{
	int f;
	char szFName[MAXPATHLEN];

	if (!blob) {
		rpc_SetErr(EINVAL, "Invalid argument BLOB");
		return -1;
	}
	if (blob->blob_data) {
		rpc_SetErr(EPERM, "Already mmapped object found!");
		return -1;
	}

	memset(szFName, 0, sizeof szFName);
	snprintf(szFName, sizeof szFName, BLOB_FILE, AIT_GET_STRZ(&srv->srv_blob.dir), blob->blob_var);
	f = open(szFName, O_RDWR);
	if (f == -1) {
		LOGERR;
		return -1;
	}

	blob->blob_data = mmap(NULL, blob->blob_len, PROT_READ | PROT_WRITE, MAP_SHARED, f, 0);
	if (blob->blob_data == MAP_FAILED) {
		LOGERR;
		close(f);
		blob->blob_data = NULL;
		return -1;
	} else {
		close(f);

		madvise(blob->blob_data, blob->blob_len, MADV_SEQUENTIAL);
	}

	return 0;
}

/*
 * rpc_srv_blobUnmap() - Unmap blob memory region 
 *
 * @blob = Mapped BLOB element
 * return: none
 */
void
rpc_srv_blobUnmap(rpc_blob_t * __restrict blob)
{
	if (blob && blob->blob_data) {
		munmap(blob->blob_data, blob->blob_len);
		blob->blob_data = NULL;
	}
}

/*
 * rpc_srv_blobFree() - Free blob from disk & memory
 *
 * @srv = RPC Server instance
 * @blob = Mapped BLOB element
 * return: -1 error or 0 ok
 */
int
rpc_srv_blobFree(rpc_srv_t * __restrict srv, rpc_blob_t * __restrict blob)
{
	char szFName[MAXPATHLEN];

	if (!blob) {
		rpc_SetErr(EINVAL, "Invalid argument BLOB");
		return -1;
	} else
		rpc_srv_blobUnmap(blob);

	schedCancelby(srv->srv_blob.root, taskTIMER, CRITERIA_ARG, blob, NULL);

	memset(szFName, 0, sizeof szFName);
	snprintf(szFName, sizeof szFName, BLOB_FILE, AIT_GET_STRZ(&srv->srv_blob.dir), blob->blob_var);
	if (unlink(szFName) == -1) {
		LOGERR;
		return -1;
	}

	return 0;
}

/* ------------------------------------------------------------ */

/*
 * rpc_srv_sendBLOB() - Send mapped BLOB to client
 *
 * @cli = Client instance
 * @blob = Mapped BLOB element
 * return: -1 error, 0 ok
 */
int
rpc_srv_sendBLOB(rpc_cli_t * __restrict cli, rpc_blob_t * __restrict blob)
{
	int ret, len;
	uint8_t *pos;

	if (!cli || !blob || !blob->blob_data) {
		rpc_SetErr(EINVAL, "Invalid arguments");
		return -1;
	}

	for (ret = blob->blob_len, pos = blob->blob_data; ret > 0; ret -= len, pos += len) {
		len = send(cli->cli_sock, pos, ret, MSG_NOSIGNAL);
		if (len == -1) {
			LOGERR;
			return -1;
		}
	}

	return ret;
}

/*
 * rpc_srv_recvBLOB() - Receive BLOB from client
 *
 * @cli = Client instance
 * @blob = Mapped BLOB element
 * return: -1 error, 0 ok, >0 unreceived data from client, may be error?
 */
int
rpc_srv_recvBLOB(rpc_cli_t * __restrict cli, rpc_blob_t * __restrict blob)
{
	int ret, len;
	uint8_t *pos;
	struct pollfd pfd;

	if (!cli || !blob || !blob->blob_data) {
		rpc_SetErr(EINVAL, "Invalid arguments");
		return -1;
	}

	pfd.fd = cli->cli_sock;
	pfd.events = POLLIN | POLLPRI;
	for (ret = blob->blob_len, pos = blob->blob_data; ret > 0; ret -= len, pos += len) {
		if ((len = poll(&pfd, 1, DEF_RPC_TIMEOUT * 1000)) < 1 || 
				pfd.revents & (POLLERR | POLLHUP | POLLNVAL)) {
			if (len)
				LOGERR;
			else
				rpc_SetErr(ETIMEDOUT, "Timeout reached! Server not respond");
			return -1;
		}

		len = recv(cli->cli_sock, pos, ret, 0);
		if (len == -1) {
			LOGERR;
			return -1;
		}
	}

	return ret;
}
