/*************************************************************************
* (C) 2010 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: lists.c,v 1.18 2015/07/02 22:28:15 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


#pragma GCC visibility push(hidden)

static int
rpc_funcs_cmp(struct tagRPCFunc *a, struct tagRPCFunc *b)
{
	int ret;

	assert(a && b);

	ret = AIT_KEY(&a->func_name) - AIT_KEY(&b->func_name);

	if (ret < 0)
		return -1;
	else if (ret > 0)
		return 1;

	return ret;
}

AVL_GENERATE(tagRPCFuncs, tagRPCFunc, func_node, rpc_funcs_cmp);

#pragma GCC visibility pop

/*
 * rpc_srv_registerCall() - Register call to RPC server
 *
 * @srv = RPC Server instance
 * @tag = Function tag
 * @funcaddr = Function address
 * return: -1 error, 0 already registered tag or 1 register ok
 */
int
rpc_srv_registerCall(rpc_srv_t * __restrict srv, u_short tag, void *funcaddr)
{
	rpc_func_t *func;

	if (!srv || !funcaddr) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t register function");
		return -1;
	}

	if (!(func = e_malloc(sizeof(rpc_func_t)))) {
		LOGERR;
		return -1;
	} else {
		memset(func, 0, sizeof(rpc_func_t));
		AIT_KEY(&func->func_name) = tag;
	}

	/* search for duplicate */
	if (AVL_FIND(tagRPCFuncs, &srv->srv_funcs, func)) {
		e_free(func);
		return 0;
	}

	func->func_parent = srv;
	AIT_SET_PTR(&func->func_name, funcaddr, 0);

	/* add to list of functions */
	RPC_FUNCS_LOCK(&srv->srv_funcs);
	SLIST_INSERT_HEAD(&srv->srv_funcs, func, func_next);
	AVL_INSERT(tagRPCFuncs, &srv->srv_funcs, func);
	RPC_FUNCS_UNLOCK(&srv->srv_funcs);
	return 1;
}

/*
 * rpc_srv_unregisterCall() - Unregister call from RPC server
 *
 * @srv = RPC Server instance
 * @tag = Function tag
 * return: -1 error, 0 not found call, 1 unregister ok
 */
int
rpc_srv_unregisterCall(rpc_srv_t * __restrict srv, u_short tag)
{
	rpc_func_t *f;

	if (!srv) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t unregister function");
		return -1;
	}

	f = rpc_srv_getCall(srv, tag);
	if (!f) 		/* not found element for unregister */
		return 0;

	RPC_FUNCS_LOCK(&srv->srv_funcs);
	AVL_REMOVE(tagRPCFuncs, &srv->srv_funcs, f);
	SLIST_REMOVE(&srv->srv_funcs, f, tagRPCFunc, func_next);
	RPC_FUNCS_UNLOCK(&srv->srv_funcs);

	AIT_FREE_VAL(&f->func_name);
	e_free(f);
	return 1;
}

/*
 * rpc_srv_getCall()  - Get registered call from RPC server
 *
 * @srv = RPC Server instance
 * @tag = tag for function
 * return: NULL not found call, !=NULL return call
 */
rpc_func_t *
rpc_srv_getCall(rpc_srv_t * __restrict srv, uint16_t tag)
{
	rpc_func_t tmp;

	if (!srv) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t get function");
		return NULL;
	} else
		memset(&tmp, 0, sizeof tmp);

	AIT_KEY(&tmp.func_name) = tag;
	return AVL_FIND(tagRPCFuncs, &srv->srv_funcs, &tmp);
}

/* --------------------------------------------------------- */

/*
 * rpc_srv_getBLOB() - Get registered BLOB 
 *
 * @srv = RPC Server instance
 * @var = hash for variable
 * return: NULL not found, !=NULL return blob var
 */
rpc_blob_t *
rpc_srv_getBLOB(rpc_srv_t * __restrict srv, uint32_t var)
{
	rpc_blob_t *b, *tmp;

	if (!srv) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t get BLOB variable");
		return NULL;
	}

	TAILQ_FOREACH_SAFE(b, &srv->srv_blob.blobs, blob_node, tmp)
		if (b->blob_var == var)
			break;

	return b;
}

/*
 * rpc_srv_registerBLOB() - Register new BLOB to server
 *
 * @srv = RPC Server instance
 * @len = BLOB length
 * @tout = BLOB live timeout in seconds
 * return: NULL error or new registered BLOB
 */
rpc_blob_t *
rpc_srv_registerBLOB(rpc_srv_t * __restrict srv, size_t len, int tout)
{
	rpc_blob_t *blob = NULL;

	if (!srv || !len) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t register BLOB variable");
		return blob;
	}

	blob = rpc_srv_blobCreate(srv, len, tout);

	TAILQ_INSERT_TAIL(&srv->srv_blob.blobs, blob, blob_node);

	return blob;
}

/*
 * rpc_srv_unregisterBLOB() - Unregister BLOB from server
 *
 * @srv = RPC Server instance
 * @var = BLOB Variable for unregister
 * return: -1 error, 0 not found call, 1 unregister ok
 */
int
rpc_srv_unregisterBLOB(rpc_srv_t * __restrict srv, uint32_t var)
{
	rpc_blob_t *b;

	if (!srv) {
		rpc_SetErr(EINVAL, "Invalid parameter can`t unregister BLOB variable");
		return -1;
	}

	b = rpc_srv_getBLOB(srv, var);
	if (!b)				/* not found element for unregister */
		return 0;

	TAILQ_REMOVE(&srv->srv_blob.blobs, b, blob_node);

	rpc_srv_blobFree(srv, b);
	e_free(b);
	return 1;
}
